@@ -1,5 +1,5 @@
 /*
-  Copyright 2007-2012 David Robillard <http://drobilla.net>
+  Copyright 2007-2016 David Robillard <http://drobilla.net>
 
   Permission to use, copy, modify, and/or distribute this software for any
   purpose with or without fee is hereby granted, provided that the above
@@ -14,22 +14,26 @@
   OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
 
-#include "lv2/lv2plug.in/ns/ext/worker/worker.h"
-
 #include "jalv_internal.h"
 
+#include "lv2/worker/worker.h"
+
 void
 jalv_worker_init(Jalv*                       jalv,
                  JalvWorker*                 worker,
-                 const LV2_Worker_Interface* iface);
+                 const LV2_Worker_Interface* iface,
+                 bool                        threaded);
 
 void
 jalv_worker_finish(JalvWorker* worker);
 
+void
+jalv_worker_destroy(JalvWorker* worker);
+
 LV2_Worker_Status
 jalv_worker_schedule(LV2_Worker_Schedule_Handle handle,
                      uint32_t                   size,
                      const void*                data);
 
 void
-jalv_worker_emit_responses(Jalv* jalv, JalvWorker* worker);
+jalv_worker_emit_responses(JalvWorker* worker, LilvInstance* instance);
