@@ -1,5 +1,5 @@
 /*
-  Copyright 2012 David Robillard <http://drobilla.net>
+  Copyright 2012-2014 David Robillard <http://drobilla.net>
 
   Permission to use, copy, modify, and/or distribute this software for any
   purpose with or without fee is hereby granted, provided that the above
@@ -17,6 +17,8 @@
 #ifndef ZIX_SEM_H
 #define ZIX_SEM_H
 
+#include "zix/common.h"
+
 #ifdef __APPLE__
 #    include <mach/mach.h>
 #elif defined(_WIN32)
@@ -24,10 +26,9 @@
 #    include <windows.h>
 #else
 #    include <semaphore.h>
+#    include <errno.h>
 #endif
 
-#include "zix/common.h"
-
 #ifdef __cplusplus
 extern "C" {
 #endif
@@ -59,13 +60,13 @@ extern "C" {
 typedef struct ZixSemImpl ZixSem;
 
 /**
-   Create and initialize @c sem to @c initial.
+   Create and initialize `sem` to `initial`.
 */
 static inline ZixStatus
 zix_sem_init(ZixSem* sem, unsigned initial);
 
 /**
-   Destroy @c sem.
+   Destroy `sem`.
 */
 static inline void
 zix_sem_destroy(ZixSem* sem);
@@ -81,7 +82,7 @@ zix_sem_post(ZixSem* sem);
    Wait until count is > 0, then decrement.
    Obviously not realtime safe.
 */
-static inline void
+static inline ZixStatus
 zix_sem_wait(ZixSem* sem);
 
 /**
@@ -103,9 +104,9 @@ struct ZixSemImpl {
 };
 
 static inline ZixStatus
-zix_sem_init(ZixSem* sem, unsigned initial)
+zix_sem_init(ZixSem* sem, unsigned val)
 {
-	return semaphore_create(mach_task_self(), &sem->sem, SYNC_POLICY_FIFO, 0)
+	return semaphore_create(mach_task_self(), &sem->sem, SYNC_POLICY_FIFO, val)
 		? ZIX_STATUS_ERROR : ZIX_STATUS_SUCCESS;
 }
 
@@ -121,10 +122,13 @@ zix_sem_post(ZixSem* sem)
 	semaphore_signal(sem->sem);
 }
 
-static inline void
+static inline ZixStatus
 zix_sem_wait(ZixSem* sem)
 {
-	semaphore_wait(sem->sem);
+	if (semaphore_wait(sem->sem) != KERN_SUCCESS) {
+		return ZIX_STATUS_ERROR;
+	}
+	return ZIX_STATUS_SUCCESS;
 }
 
 static inline bool
@@ -159,16 +163,19 @@ zix_sem_post(ZixSem* sem)
 	ReleaseSemaphore(sem->sem, 1, NULL);
 }
 
-static inline void
+static inline ZixStatus
 zix_sem_wait(ZixSem* sem)
 {
-	WaitForSingleObject(sem->sem, INFINITE);
+	if (WaitForSingleObject(sem->sem, INFINITE) != WAIT_OBJECT_0) {
+		return ZIX_STATUS_ERROR;
+	}
+	return ZIX_STATUS_SUCCESS;
 }
 
 static inline bool
 zix_sem_try_wait(ZixSem* sem)
 {
-	WaitForSingleObject(sem->sem, 0);
+	return WaitForSingleObject(sem->sem, 0) == WAIT_OBJECT_0;
 }
 
 #else  /* !defined(__APPLE__) && !defined(_WIN32) */
@@ -196,14 +203,17 @@ zix_sem_post(ZixSem* sem)
 	sem_post(&sem->sem);
 }
 
-static inline void
+static inline ZixStatus
 zix_sem_wait(ZixSem* sem)
 {
-	/* Note that sem_wait always returns 0 in practice, except in
-	   gdb (at least), where it returns nonzero, so the while is
-	   necessary (and is the correct/safe solution in any case).
-	*/
-	while (sem_wait(&sem->sem) != 0) {}
+	while (sem_wait(&sem->sem)) {
+		if (errno != EINTR) {
+			return ZIX_STATUS_ERROR;
+		}
+		/* Otherwise, interrupted, so try again. */
+	}
+
+	return ZIX_STATUS_SUCCESS;
 }
 
 static inline bool
